package saml

import (
	"crypto/rand"
	"crypto/tls"
	"encoding/base64"
	"encoding/hex"
	"encoding/xml"
	"errors"
	"fmt"
	"io"
	"io/ioutil"
	"log"
	"net/http"
	"net/url"
	"os"
	"strings"
	"time"

	"github.com/crewjam/saml"
	"github.com/crewjam/saml/logger"
	"github.com/gorilla/mux"
	"gopkg.in/yaml.v2"

	"git.autistici.org/id/go-sso/httpsso"
)

type serviceProvider struct {
	// Descriptor can either be an inline XML document, or it can
	// be read from a file with the syntax "@filename".
	Descriptor string   `yaml:"descriptor"`
	SSOGroups  []string `yaml:"sso_groups"`

	parsed *saml.EntityDescriptor
}

type Config struct {
	BaseURL string `yaml:"base_url"`

	UsersFile string `yaml:"users_file"`

	// SAML X509 credentials.
	CertificateFile string `yaml:"certificate_file"`
	PrivateKeyFile  string `yaml:"private_key_file"`

	// SSO configuration.
	SessionAuthKey    string `yaml:"session_auth_key"`
	SessionEncKey     string `yaml:"session_enc_key"`
	SSOLoginServerURL string `yaml:"sso_server_url"`
	SSOPublicKeyFile  string `yaml:"sso_public_key_file"`
	SSODomain         string `yaml:"sso_domain"`

	// Service provider config.
	ServiceProviders   []*serviceProvider `yaml:"service_providers"`
	serviceProviderMap map[string]*serviceProvider
}

// Sanity checks for the configuration.
func (c *Config) check() error {
	switch len(c.SessionAuthKey) {
	case 32, 64:
	case 0:
		return errors.New("session_auth_key is empty")
	default:
		return errors.New("session_auth_key must be a random string of 32 or 64 bytes")
	}
	switch len(c.SessionEncKey) {
	case 16, 24, 32:
	case 0:
		return errors.New("session_enc_key is empty")
	default:
		return errors.New("session_enc_key must be a random string of 16, 24 or 32 bytes")
	}
	if c.SSOLoginServerURL == "" {
		return errors.New("sso_server_url is empty")
	}
	if c.SSODomain == "" {
		return errors.New("sso_domain is empty")
	}
	return nil
}

func (c *Config) loadServiceProviders() error {
	c.serviceProviderMap = make(map[string]*serviceProvider)
	for _, sp := range c.ServiceProviders {
		var data []byte
		if strings.HasPrefix(sp.Descriptor, "@") {
			var err error
			data, err = ioutil.ReadFile(sp.Descriptor[1:])
			if err != nil {
				return fmt.Errorf("reading %s: %v", sp.Descriptor[1:], err)
			}
		} else {
			data = []byte(sp.Descriptor)
		}
		var ent saml.EntityDescriptor
		if err := xml.Unmarshal(data, &ent); err != nil {
			return fmt.Errorf("error unmarshaling descriptor: %v", err)
		}
		sp.parsed = &ent
		c.serviceProviderMap[ent.EntityID] = sp
	}
	return nil
}

func (c *Config) GetServiceProvider(r *http.Request, serviceProviderID string) (*saml.EntityDescriptor, error) {
	sp, ok := c.serviceProviderMap[serviceProviderID]
	if !ok {
		return nil, os.ErrNotExist
	}
	return sp.parsed, nil
}

func (c *Config) GetSSOGroups(serviceProviderID string) []string {
	sp, ok := c.serviceProviderMap[serviceProviderID]
	if !ok {
		return nil
	}
	return sp.SSOGroups
}

// Read users from a YAML-encoded file, in a format surprisingly
// compatible with git.autistici.org/id/auth/server.
//
// TODO: Make it retrieve the email addresses as extra data in the SSO
// token (this feature is currently unsupported by the SSO server,
// even though the auth-server provides the information).
type userInfo struct {
	Name  string `yaml:"name"`
	Email string `yaml:"email"`
}

type userFileBackend struct {
	config *Config
	users  map[string]userInfo
}

func newUserFileBackend(config *Config) (*userFileBackend, error) {
	data, err := ioutil.ReadFile(config.UsersFile)
	if err != nil {
		return nil, err
	}
	var userList []userInfo
	if err := yaml.Unmarshal(data, &userList); err != nil {
		return nil, err
	}
	users := make(map[string]userInfo)
	for _, u := range userList {
		users[u.Name] = u
	}
	return &userFileBackend{
		config: config,
		users:  users,
	}, nil
}

// Nice little O(N^2) algorithm right there...
func matchGroups(user, exp []string) bool {
	if exp == nil {
		return true
	}
	for _, ug := range user {
		for _, eg := range exp {
			if ug == eg {
				return true
			}
		}
	}
	return false
}

func (b *userFileBackend) GetSession(w http.ResponseWriter, r *http.Request, req *saml.IdpAuthnRequest) *saml.Session {
	// Check for authentication by verifying the SSO username. We
	// also need to be able to retrieve user information from the
	// backend, to match SSO. Group membership, if enabled in our
	// configuration, is also verified at this stage.
	username := httpsso.Username(r)
	if username == "" {
		http.Error(w, "No user found", http.StatusInternalServerError)
		return nil
	}

	if !matchGroups(httpsso.Groups(r), b.config.GetSSOGroups(req.ServiceProviderMetadata.ID)) {
		http.Error(w, "Forbidden (bad group)", http.StatusForbidden)
		return nil
	}

	user, ok := b.users[username]
	if !ok {
		http.Error(w, "User not found", http.StatusInternalServerError)
		return nil
	}

	log.Printf("successfully authenticated session for username=%s, provider=%s", username, req.ServiceProviderMetadata.ID)

	return &saml.Session{
		ID:             base64.StdEncoding.EncodeToString(randomBytes(32)),
		CreateTime:     saml.TimeNow(),
		ExpireTime:     saml.TimeNow().Add(sessionMaxAge),
		Index:          hex.EncodeToString(randomBytes(32)),
		UserName:       user.Name,
		UserEmail:      user.Email,
		UserCommonName: user.Name,
		UserGivenName:  user.Name,
	}
}

func NewSAMLIDP(config *Config) (http.Handler, error) {
	if err := config.check(); err != nil {
		return nil, err
	}
	if err := config.loadServiceProviders(); err != nil {
		return nil, err
	}

	cert, err := tls.LoadX509KeyPair(config.CertificateFile, config.PrivateKeyFile)
	if err != nil {
		return nil, err
	}

	pkey, err := ioutil.ReadFile(config.SSOPublicKeyFile)
	if err != nil {
		return nil, err
	}

	w, err := httpsso.NewSSOWrapper(config.SSOLoginServerURL, pkey, config.SSODomain, []byte(config.SessionAuthKey), []byte(config.SessionEncKey))
	if err != nil {
		return nil, err
	}

	baseURL, err := url.Parse(config.BaseURL)
	if err != nil {
		return nil, err
	}
	ssoURL := *baseURL
	ssoURL.Path += "/login/"
	metadataURL := *baseURL
	metadataURL.Path += "/metadata"
	svc := fmt.Sprintf("%s%s", ssoURL.Host, ssoURL.Path)
	if !strings.HasSuffix(svc, "/") {
		svc += "/"
	}

	users, err := newUserFileBackend(config)
	if err != nil {
		return nil, err
	}

	// Create the SAML IdentityProvider, but then we put another
	// mux.Router in front in order to wrap just the ssoURL with
	// our own SSO handler.
	idp := &saml.IdentityProvider{
		Key:                     cert.PrivateKey,
		Certificate:             cert.Leaf,
		Logger:                  logger.DefaultLogger,
		MetadataURL:             metadataURL,
		SSOURL:                  ssoURL,
		ServiceProviderProvider: config,
		SessionProvider:         users,
	}
	h := idp.Handler()

	root := mux.NewRouter()
	root.HandlePrefix(ssoURL.Path, w.Wrap(h, svc, nil))
	root.Handle(metadataURL.Path, h)
	return root, nil
}

func randomBytes(n int) []byte {
	b := make([]byte, n)
	if _, err := io.ReadFull(rand.Reader, b[:]); err != nil {
		panic(err)
	}
	return b
}

var sessionMaxAge = 300 * time.Second
